# https://replit.com/@vyastreb/1D-Reynolds-Solver?v=1
"""
1D simple finite difference solver for incompressible Reynolds equation for tangetial motion.
V.A. Yastrebov, MINES Paris - PSL, CNRS, Evry/Paris, France, 2024.
Assited by Copilot and GPT4.
License: BSD 3 clause
"""

import numpy as np
import matplotlib.pyplot as plt


def y(x,slope):
    return -slope*x
def dydx(x,slope):
    delta = 1e-6
    return (y(x+delta,slope)-y(x-delta,slope))/(2*delta)
def analyticalSolution(x,h0,l,slope,mu,v):
    X = x/l
    s = l*slope
    H0 = h0/s
    Pref =  6*X*(1-X) / ( (H0 + 1 - X)**2 * (1 + 2*H0) )
    p = Pref * mu * v * l  / (s**2)
    return p

# Physical parameters
h0 = .002
x_min = 0
x_max = 1
l = x_max - x_min
mu = 1.
V0 = 1.
slope = 0.01

## Finite difference parameters
# Number of elements
N = 30
# Discretization step
dx = (x_max - x_min)/N
# Discretized x space
x = np.linspace(x_min, x_max, N)
# For visualization needs, plots every ith point
plot_every = 1


H = y(x,slope) - np.min(y(x,slope)) + h0
dHdx = dydx(x,slope)
print(np.min(H),np.max(H))
print('s=',slope)
H3 = H**3
H2 = H**2

matrix = np.zeros((N,N))
rhs = np.zeros(N)
for i in range(N):
    for j in range(N):
        if i == j:
            matrix[i,j] = -3*H2[i]*dHdx[i]/dx - 2*H3[i]/dx**2
        elif i == j-1:
            matrix[i,j] =  3*H2[i]*dHdx[i]/dx + H3[i]/dx**2
        elif i == j+1:
            matrix[i,j] = H3[i]/dx**2
        else:
            matrix[i,j] = 0
        rhs[i] = 6*mu*V0*dHdx[i]

## Impose Dirichlet boundary conditions of zero pressure at the inlet and outlet
# Set penalty much larger than the maximal value of the matrix       
penalty = 1e5*np.max(np.abs(matrix))
# Impose penalty coefficient on the diagonal terms
matrix[0,0]= penalty
matrix[-1,-1] = penalty
# Zero right hand side for these indices
rhs[0] = 0
rhs[-1] = 0

# Solve the linear system
P = np.linalg.solve(matrix, rhs)

# Visualization
xa = np.linspace(x_min,x_max,100)

fig, ax = plt.subplots()
plt.grid()
plt.xlim(x_min,x_max)
plt.ylim(0,1.1*np.max(P))
plt.plot(x[::plot_every], P[::plot_every],"o",color="#6666ff",label = "Numerical solution")
plt.plot(xa,analyticalSolution(xa,h0,l,slope,mu,V0),"-",color="k",label="Analytical solution")
plt.xlabel('x')
plt.ylabel('Pressure')
plt.title('Pressure Distribution')
plt.legend(loc="best")
fig.savefig("Pressure_h0_{0:.3e}_slope_{1:.3e}.pdf".format(h0,slope))
plt.show()



