"""
2D simple finite difference solver for incompressible Reynolds equation for tangetial motion.
V.A. Yastrebov, MINES Paris - PSL, CNRS, Evry/Paris, France, 2024.
Assited by Copilot and GPT4.
License: BSD 3 clause
"""

import numpy as np
import matplotlib.pyplot as plt
import sys
from scipy.sparse import diags
from scipy.sparse.linalg import spsolve

def z(x, y, slope, y0, curvature):
    # Inclined plane
    z = curvature*(y-y0)**2 - slope*x 
    return z
def dzdx(x,y, slope, y0, curvature):
    delta = 1e-6
    return (z(x+delta,y, slope, y0, curvature)-z(x-delta,y, slope, y0, curvature))/(2*delta)
def dzdy(x,y, slope, y0, curvature):
    delta = 1e-6
    return (z(x,y+delta, slope, y0, curvature)-z(x,y-delta, slope, y0, curvature))/(2*delta)

# Physical parameters
ProblemType = "lubrication"
mu = 1.0  # fluid viscosity
V0 = 1.   # sliding velocity
h0 = .2 # minimal clearance
slope = 0.2
# Define the domain
x_min = y_min = 0
x_max = y_max = 1
R = (x_max - x_min)/2.2 # Radius of the slider
curvature = slope*0.1
y0 = (y_max-y_min)/2.

# Computational parameters
N = 200
x = np.linspace(x_min, x_max, N)
y = np.linspace(y_min, y_max, N)
dx = x[1] - x[0]
X,Y = np.meshgrid(x, y)

# Compute film thickness h(x, y)
zd = z(X, Y, slope, y0, curvature)
H = zd - np.min(zd) + h0


H3 = H**3
H2 = H**2
dHdx = dzdx(X,Y, slope, y0, curvature)
dHdy = dzdy(X,Y, slope, y0, curvature)
# For simplicity precompute some matrices
term1 = 3*H2*(dHdx+dHdy)/dx + 4*H3/dx**2 
term2x = 3*H2*dHdx/dx + H3/dx**2
term2y = 3*H2*dHdy/dx + H3/dx**2

# Construction of 5-diagonal matrix
Cup =   np.zeros(N**2 - 1)
Cdown = np.zeros(N**2 - 1)
Bup   = np.zeros(N**2 - N)
Bdown = np.zeros(N**2 - N)
A = np.zeros(N**2)
max_term = 0 # maximal diagonal component of the matrix
for i in range(N):
    for j in range(N):
        index = j + i*N
        if index < N**2 - 1:
            Cup[index] = term2x[i,j]
        if index > 0:
            Cdown[index-1] = H3[i,j] / dx**2
        if index < N**2 - N:
            Bup[index] = term2y[i,j]
        if index >= N:
            Bdown[index-N] = H3[i,j] / dx**2
        A[index] = -term1[i,j]
        if abs(A[index]) > max_term:
            max_term = abs(A[index])

offsets = np.array([-N, -1, 0, 1, N])
FDmatrix = diags([Bdown, Cdown, A, Cup, Bup], offsets, shape=(N**2,N**2), format='csr')
print("// SYSTEM: Size of sparse format:", sys.getsizeof(FDmatrix)/1024/1024, "Mb")

# Compare to the full format
# FDmatrixs = FDmatrix.toarray()
# print("Size of full format:", sys.getsizeof(FDmatrixs)/1024/1024, "Mb")

# Right-hand side vector
rhs = 6 * mu * V0 * dHdx.flatten()

penalty = 1e6*max_term
if ProblemType == "lubrication":
# In lubrication problem, we apply trivial Dirichlet boundary conditions on the contour of the bearing
    for i in range(N):
        for j in range(N):
            index = j + i*N
            # Circular domain
            if (i-N/2)**2 + (j-N/2)**2 >= (R/dx)**2:
                FDmatrix[index, index] = penalty
                rhs[index] = 0
                continue
elif ProblemType == "leackage":
# In pressure driven leackage/sealing problem, we apply nontrivial Dirichlet boundary conditions: inlet and outlet pressures

    Pinlet = 1.
    Poutlet = 0
    for i in range(N):
        for j in range(N):
            index = j + i*N
            # Square domain
            if i == 0:
                FDmatrix[index, index] = penalty
                rhs[index] = penalty*Pinlet
                continue
            if i == N-1:
                FDmatrix[index, index] = penalty
                rhs[index] = penalty*Poutlet
                continue
            

# Solve the sparse linear system
import time
time0 = time.time()
# P = np.linalg.solve(M, rhs)
P = spsolve(FDmatrix, rhs)
P = P.reshape(N, N)
time1 = time.time()
print("// SYSTEM: Solving time:", time1-time0)

# Visualization
#---------------------#
# Plot fluid pressure #
#---------------------#
fig, ax = plt.subplots()
plt.imshow(P, extent=[x_min, x_max, y_min, y_max], origin='lower',interpolation='bicubic', cmap=plt.cm.inferno)
plt.colorbar(label='Pressure')
plt.xlabel('x')
plt.ylabel('y')
plt.title('Pressure Distribution')
# To save pressure figure, uncomment this line
# fig.savefig("Pressure_h0_{0:.3f}_slope_{1:.3f}.png".format(h0,slope), dpi=300)
plt.show()

#---------------------#
# Show pressure in 3D #
#---------------------#

from mpl_toolkits.mplot3d import Axes3D
fig,ax = plt.subplots(1,1,subplot_kw={'projection':'3d'})
surf = ax.plot_surface(X, Y, P, cmap=plt.cm.inferno, linewidth=0, antialiased=True, rstride=5, cstride=5)
ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('Pressure')
plt.show()

#------------------#
# Plot streamlines #
#------------------#

V,U = -H3 * np.gradient(P, dx, dx)/(12*mu)
# clean up too small velocities to avoid plotting artifacts
V[np.abs(V)<1e-6] = 0
U[np.abs(U)<1e-6] = 0
speed = np.sqrt(U**2 + V**2)
fig,ax = plt.subplots()
# Create the streamplot
strm = plt.streamplot(X, Y, U, V, density=3, color=speed, linewidth=1.5, cmap=plt.cm.inferno)

# Add a color bar
plt.colorbar(strm.lines, label='Speed')

# Set aspect of the plot to be equal
ax.set_aspect('equal', adjustable='box')

plt.title('Streamline Plot')
plt.xlabel('X-axis')
plt.ylabel('Y-axis')
# To save streamline figure, uncomment this line
# fig.savefig('streamline_plot.pdf')
plt.show()

